;		Accelerometer Servo Leveler
;		PIC 16F627 series Micro
;		Designed by Mike Berg
;		Two Channels PWM Output
;		ADXL202 Accelerometer
;		EEPROM Reference Storage
;		6/10/04		Tested Functional
;------------------------------------------------------------
        LIST    P=16F627
	ERRORLEVEL	-302
        include <P16F627.INC>
        __CONFIG        h'3f21'	
;-------------------------------------------------------------
;               Port Set-up
;           0   1   2   3   4   5   6   7       bit
; port  A   Out Out Out Out In  X   X   X
;       B   In  In  Out Out Out Out Out Out
;------------------------------------------------------------
	ORG	0			; start code
	goto	main
	ORG	4
	goto	int_service
;-------------------------------------------------------------
        cblock  0x20
		temp_W			; 20 Interrupt buffer
		temp_S			; 21 Interrupt buffer
;
                accel_x    		; 22 Accelerometer X axis
                accel_y                 ; 23 Accelerometer Y axis
;
                accel_xcal              ; 24 Accelerometer X axis reference
                accel_ycal              ; 25 Accelerometer Y axis reference
;
                xservo_out              ; 26 X axis sero PWM out
                yservo_out              ; 27 Y axis sero PWM out
;
                count_1                 ; 28 counter
                delay_count             ; 29 counter
		buffer_in		; 2a Input port buffer
		buffer_out		; 2b Output port buffer
;
		ee_adr			; 2c EEMEM adress
		ee_data			; 2d EEMEM data
        endc
;------------------------------------------------------------
;	PORTA,0/1 servo x/y out
;	PORTA,4 read/store accelerometer position switch
;	PORTB,0/1 accelerometer x/y in
;------------------------------------------------------------
#define X_ACC           buffer_in,0	;accel X in
#define Y_ACC           buffer_in,1	;accel Y in
;
#define	SERVO_X		PORTA,0		;X out
#define	SERVO_Y		PORTA,1		;Y out
#define	SWITCH		PORTA,4		;eeprom save / normal
#define	LED		PORTB,7		;LED out
;
#define	XBIT		buffer_out,0	;PORTA X axis
#define	YBIT		buffer_out,1	;PORTA Y axis
;
BLANK_PULSE	equ	0xec78		;20 mS between servo pulses
;
;------------------------------------------------------------
main
        clrf    PORTA
        clrf    PORTB
        movlw   0x07
        movwf   CMCON                   ;compare module off
        bsf     STATUS,RP0              ;switch to bank 1
        movlw   b'00010000'             ;port A in/outputs
        movwf   TRISA
        movlw   b'00000011'             ;port B in/outputs
        movwf   TRISB
        movlw   b'11000100'
        movwf   OPTION_REG              ;TMR0/32
	movlw	b'00000001'
	movwf	PIE1  			;TMR1 int enabled
;
        bcf     STATUS,RP0              ;bank 0
;
	movlw	b'00100001'		;T1/4, on
	movwf	T1CON			;TMR1 setup
	movlw	b'01000000'
        movwf   INTCON                  ;PIE enabled
;------------------------------------------------------------
	movlw	d'127'
	movwf	xservo_out
	movwf	yservo_out		;center servos
;
	bsf	INTCON,GIE		;enable interrupts
first_load
	bcf	LED			;LED off
	call	delay_256		;wait 256 mS
	bsf	LED			;LED on
	call	read_accel		;get accelerometer reading
	btfsc	SWITCH			;PORTA:4
	goto	eeretrieve_level	;hi selects mem read, low mem store
;------------------------------------------------------------
;	Assuming the platform with accelerometer is level we measure the
;	pulse widths of the X & Y axis accelerometers and store for
;	reference points. The first symbol stored (*) indicates valid data
;	and allows routine to run.  First time operation requires sensor to
;	be leveled and SWITCH set in order to capture and store a reference.
;
eestore_level
	bcf	INTCON,GIE
	btfsc	INTCON,GIE
	goto	$-2			;ints off
	clrf	ee_adr			;EEMEM address 0
	movlw	"*"
	movwf	ee_data
	call	ee_write		;valid data symbol
	incf	ee_adr,F		;EEMEM address 1
	movfw	accel_x			;x axis 
	movwf	ee_data			;to EEMEM
	call	ee_write
	incf	ee_adr,F		;EEMEM address 2
	movfw	accel_y			;x axis 
	movwf	ee_data			;to EEMEM
	call	ee_write
	bsf	INTCON,GIE		;enable interrupts again
;------------------------------------------------------------
;	This routine reads the previously stored reference values and copies
;	them to working registers. 
;
eeretrieve_level			;read level from EEMEM
	clrf	ee_adr			;EEMEM address 0
	call	ee_read
	movfw	ee_data
	xorlw	"*"			;see a (*)?
	bnz	first_load		;no, invalid data in eemem
	incf	ee_adr,F		;yes, read address 1
	call	ee_read			;read stored X axis
;
	movfw	ee_data
	movwf	accel_xcal		;stored X axis ref
	incf	ee_adr,F		;EEMEM address 2
;
	call	ee_read			;read stored Y axis
	movfw	ee_data			;at EEMEM address 3
	movwf	accel_ycal		;stored Y axis ref
;
;------------------------------------------------------------
;
start_here
	bcf	LED			;LED off
        call    read_accel		;read accelerometer inputs
	call	calc_accel		;adjust output servos to match
	goto	start_here		;loop back and check control
;
;------------------------------------------------------------
int_service
        movwf   temp_W			;store W and STATUS regs
        swapf   STATUS,W
        movwf   temp_S
	bcf	STATUS,RP0		;insure we're in bank 0
;
	btfsc	PIR1,TMR1IF		;test for TMR1 interrupt
	call	T1_handler		;flag set, call routine
;
outta_here
        swapf   temp_S,W		;restore W and STATUS regs
        movwf   STATUS
        swapf   temp_W,F
        swapf   temp_W,W
        retfie                          ;return to program
;------------------------------------------------------------
;		Subroutines
;------------------------------------------------------------
;	Reads PORTA to see which if either X or Y servos
;	are pulsing. If neither, it sets X and the X timeout then exits. 
;	If X is high it clears X and sets Y and the Y timeout then exits.
;	If Y is high it clears Y and sets the gap timeout to 20 ms
;
T1_handler
	bcf	PIR1,TMR1IF		;clear int flag
	movfw	PORTA
	movwf	buffer_out		;PORTA to buffer_out
	btfss	XBIT			;test X axis set
	goto	exam_y			;
	bcf	SERVO_X			;X pulse lo
	movlw	0xfe
	movwf	TMR1H			;load 1 mS delay
	comf	yservo_out,W		;load Y varible
	movwf	TMR1L			;add between 0-1 mS to delay
	bsf	SERVO_Y			;Y pulse start
	return
exam_y
	btfss	YBIT			;test Y axis input set
	goto	gap_over
	bcf	SERVO_Y			;Y pulse lo
	movlw	high BLANK_PULSE
	movwf	TMR1H
	movlw	low BLANK_PULSE		;20 mS til next X,Y pulses
	movwf	TMR1L
	return
gap_over
	movlw	0xfe
	movwf	TMR1H			;load 1 mS delay
	comf	xservo_out,W		;load X varible
	movwf	TMR1L			;add between 0-1 mS to delay
	bsf	SERVO_X			;X pulse start
	return	
;------------------------------------------------------------
;               Reads PORTB X/Y Accelerometer inputs
read_accel
	call	load_buffer_in		;buffers PORTB
        btfsc   X_ACC			;wait for low X pulse from
        goto    $-2			;accelerometer
	call	load_buffer_in
        btfss   X_ACC			;wait for hi X pulse
        goto    $-2
        clrf    TMR0			;start timing with TMR0
	call	load_buffer_in
        btfsc   X_ACC			;wait for X pulse end
        goto    $-2
        movfw   TMR0			;done timing X axis
        movwf   accel_x			;store results in accel_x
;
	call	load_buffer_in		;buffers PORTB
        btfsc   Y_ACC			;wait for low Y pulse from
        goto    $-2			;accelerometer
	call	load_buffer_in
        btfss   Y_ACC			;wait for hi Y pulse
        goto    $-2
        clrf    TMR0			;start timing with TMR0
	call	load_buffer_in
        btfsc   Y_ACC			;wait for Y pulse end
        goto    $-2
        movfw   TMR0			;done timing Y axis
        movwf   accel_y			;store results in accel_y
	return
;------------------------------------------------------------
;		Adjust servo outputs to accelerometer inputs
calc_accel
	bsf	LED
	movfw	accel_x			;compare latest accelerometer
	subwf	accel_xcal,W		;pulse width to stored reference
	bz	check_y_accel
	bnc	x_accel_over
x_accel_under
	incfsz	xservo_out,W		;X axis < than ref 
	incf	xservo_out,F		;increase X servo
	goto	check_y_accel	
x_accel_over
	decfsz	xservo_out,W		;X axis > than ref
	decf	xservo_out,F		;decrease X servo
;
check_y_accel
	movfw	accel_y
	subwf	accel_ycal,W
	bz	accel_done
	bnc	y_accel_over
y_accel_under
	incfsz	yservo_out,W		;Y axis < than ref
	incf	yservo_out,F		;increase Y servo
	goto	accel_done	
y_accel_over
	decfsz	yservo_out,W		;Y axis > than ref
	decf	yservo_out,F		;decrease Y servo
accel_done
	bcf	LED
	return
;
;------------------------------------------------------------
load_buffer_in
	movfw	PORTB			;PORTB to buffer register
	movwf	buffer_in
	return
;------------------------------------------------------------
ee_write
	movlw	EEADR			;EEMEM write routine
	movwf	FSR
	movfw	ee_adr
	movwf	INDF
	movlw	EEDATA
	movwf	FSR
	movfw	ee_data
	movwf	INDF
;
	bsf	STATUS,RP0
	bsf	EECON1,WREN
	movlw	0x55
	movwf	EECON2
	movlw	0xaa
	movwf	EECON2
	bsf	EECON1,WR
	btfsc	EECON1,WR
	goto	$-1
	bcf	EECON1,WREN
	bcf	STATUS,RP0
	return
ee_read
	movlw	EEADR			;EEMEM read routine
	movwf	FSR
	movfw	ee_adr
	movwf	INDF
	bsf	STATUS,RP0
	bsf	EECON1,RD
	bcf	STATUS,RP0
	movlw	EEDATA
	movwf	FSR
	movfw	INDF
	movwf	ee_data	
	return
;------------------------------------------------------------
delay_256
        movlw   0xff                    ;256 mS delay
        movwf   delay_count
	movwf	count_1
delay_loop
	nop
	decfsz	count_1,F
	goto	delay_loop
        decfsz  delay_count,F
        goto    delay_loop
        return
;------------------------------------------------------------
        end
;------------------------------------------------------------
Copyright (c) 2004, Mike Berg
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions 
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer. 
Redistributions in binary form must reproduce the above copyright 
notice, this list of conditions and the following disclaimer in the 
documentation and/or other materials provided with the distribution. 
Neither the name of the <ORGANIZATION> nor the names of its 
contributors may be used to endorse or promote products derived from 
this software without specific prior written permission. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.

